package org.postgresql.core;

/**
 * UTF8Encoder is here as an optimization for chat-to-UTF8 encoding,
 * to replace the slow standard code from Java up to 1.3.1 (versions
 * starting from 1.4.0 perform significantly better at it). This code
 * was adapted from ftp://www.unicode.org/Public/PROGRAMS/CVTUTF/.
 *
 * <pre>
 * Copyright 2001 Unicode, Inc.
 * 
 * Disclaimer
 * 
 * This source code is provided as is by Unicode, Inc. No claims are
 * made as to fitness for any particular purpose. No warranties of any
 * kind are expressed or implied. The recipient agrees to determine
 * applicability of information provided. If this file has been
 * purchased on magnetic or optical media from Unicode, Inc., the
 * sole remedy for any claim will be exchange of defective media
 * within 90 days of receipt.
 * 
 * Limitations on Rights to Redistribute This Code
 * 
 * Unicode, Inc. hereby grants the right to freely use the information
 * supplied in this file in the creation of products supporting the
 * Unicode Standard, and to make copies of this file in any form
 * for internal or external distribution as long as this notice
 * remains attached.
 * </pre>
 */
public class UTF8Encoder
{
    private static final int UNI_REPLACEMENT_CHAR = 0x0000FFFD;
    private static final int UNI_SUR_HIGH_START = 0xD800;
    private static final int UNI_SUR_HIGH_END = 0xDBFF;
    private static final int UNI_SUR_LOW_START = 0xDC00;
    private static final int UNI_SUR_LOW_END = 0xDFFF;

    /**
     * Once the bits are split out into bytes of UTF-8, this is a mask OR-ed
     * into the first byte, depending on how many bytes follow.  There are
     * as many entries in this table as there are UTF-8 sequence types.
     * (I.e., one byte sequence, two byte... six byte sequence.)
     */
    private static final int FIRST_BYTE_MARK[] = { 0x00, 0x00, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC };

    private static final int HALF_SHIFT = 10; /* used for shifting by 10 bits */
    private static final int HALF_BASE = 0x0010000;

    public static final int STRICT_CONVERSION = 0;
    public static final int LENIENT_CONVERSION = 1;
    public static final int CONVERSION_OK = 0;
    public static final int SOURCE_EXHAUSTED = 1;
    public static final int TARGET_EXHAUSTED = 2;
    public static final int SOURCE_ILLEGAL = 3;

    public static void main(String[] args) throws Exception
    {
        // Test #1: encode all Unicode characters,
        // compare against results from the standard Java encoder

        char[] in = new char[1];
        
        for (char i = 0; i < Character.MAX_VALUE; i++)
        {
            String str = String.valueOf(in[0]);
            byte[] b = str.getBytes("UTF-8");
            byte[] buf1 = encode1(str);
            byte[] buf2 = encode2(str);

            if (b.length != buf1.length || b.length != buf2.length) fail(str);

            for (int j = 0; j < b.length; j++)
                if (b[j] != buf1[j] || b[j] != buf2[j]) fail(str);
        }

        // Test #2: encode 10000 random Strings,
        // compare against results from the standard Java encoder

        java.util.Random rnd = new java.util.Random();

        for (int i = 0; i < 10000; i++)
        {
            int len = Math.abs(rnd.nextInt()) % 1000;
            StringBuffer buf = new StringBuffer();
            
            for (int j = 0; j < len; j++)
            {
                char c = (char) (Math.abs(rnd.nextInt()) % 0xFFFE);

                if (c >= UNI_SUR_LOW_START && c <= UNI_SUR_LOW_END)
                    continue;
                else if (c >= UNI_SUR_HIGH_START && c <= UNI_SUR_HIGH_END)
                {
                    buf.append(c);
                    c = (char) ((Math.abs(rnd.nextInt()) %   
                        (UNI_SUR_LOW_END - UNI_SUR_LOW_START)) +
                        UNI_SUR_LOW_START);

                    buf.append(c);
                }
                else buf.append(c);
            }
            String str = buf.toString();
            
            byte[] b1 = encode1(str);
            byte[] b2 = encode2(str);
            byte[] b3 = buf.toString().getBytes("UTF-8");

            if (b1.length != b2.length || b2.length != b3.length)
                fail(str);

            for (int j = 0; j < b1.length; j++)
                if (b1[j] != b2[j] || b2[j] != b3[j]) fail(str);
        }
    }

    private static void fail(String str) throws Exception
    {
        // report test failure

        StringBuffer strb = new StringBuffer();
        for (int k = 0; k < str.length(); k++)
        {
            strb.append(Integer.toHexString((int) str.charAt(k)));
            strb.append(' ');
            throw new Exception("test failed on String: " + strb);
        }
    }

    /**
     * Encodes a String into UTF-8. This implementation grows the output
     * buffer dynamically, which may result in copying data around the
     * memory and increased garbage collection.
     */
    public static byte[] encode1(String s)
        throws Exception
    {
        final int BYTE_MASK = 0xBF;
        final int BYTE_MARK = 0x80;
        final int flags = STRICT_CONVERSION;

        char[] source = s.toCharArray();
        int sPtr = 0, sEnd = source.length;

        // We don't know how many bytes the target buffer will need at this time.
        // We know that it will be a number in range [source.length, source.length*4]
        // Let's assume 5% of non-ASCII characters in the input and grow the
        // buffer if it turns out to have been a bad guess.
        
        byte[] target = new byte[(int)(sPtr * 1.05f)];
        int tPtr = 0, tEnd = target.length;
        
        while (sPtr < sEnd)
        {
            int bytesToWrite = 0;
            int ch = source[sPtr++];
            
            /* If we have a surrogate pair, convert to UTF32 first. */
            
            if (ch >= UNI_SUR_HIGH_START && ch <= UNI_SUR_HIGH_END && sPtr < sEnd)
            {
                int ch2 = source[sPtr];

                if (ch2 >= UNI_SUR_LOW_START && ch2 <= UNI_SUR_LOW_END)
                {
                    ch = ((ch - UNI_SUR_HIGH_START) << HALF_SHIFT)
                        + (ch2 - UNI_SUR_LOW_START) + HALF_BASE;
                    ++sPtr;
                }
                else if (flags == STRICT_CONVERSION)
                {
                    /* it's an unpaired high surrogate */
                    --sPtr; /* return to the illegal value itself */
                    throw new Exception("expected a surrogate character at position " + sPtr);
                }
            }
            else if ((flags == STRICT_CONVERSION) && (ch >= UNI_SUR_LOW_START && ch <= UNI_SUR_LOW_END))
            {
                --sPtr; /* return to the illegal value itself */
                throw new Exception("expected a surrogate character at position " + sPtr);
            }
            
            /* Figure out how many bytes the result will require */
            
            if (ch < 0x80) bytesToWrite = 1;
            else if (ch < 0x800) bytesToWrite = 2;
            else if (ch < 0x10000) bytesToWrite = 3;
            else if (ch < 0x200000) bytesToWrite = 4;
            else
            {
                bytesToWrite = 2;
                ch = UNI_REPLACEMENT_CHAR;
            }

            tPtr += bytesToWrite;
            
            if (tPtr > tEnd)
            {
                byte[] targetOld = target;
                target = new byte[Math.min(tPtr * 2, source.length * 4)];
                System.arraycopy(targetOld, 0, target, 0, tPtr-bytesToWrite);
                tEnd = target.length;
            }
            switch (bytesToWrite)
            {    /* note: everything falls through. */
                case 4:    target[--tPtr] = (byte)((ch | BYTE_MARK) & BYTE_MASK); ch >>= 6;
                case 3:    target[--tPtr] = (byte)((ch | BYTE_MARK) & BYTE_MASK); ch >>= 6;
                case 2:    target[--tPtr] = (byte)((ch | BYTE_MARK) & BYTE_MASK); ch >>= 6;
                case 1:    target[--tPtr] = (byte)(ch | FIRST_BYTE_MARK[bytesToWrite]);
            }
            tPtr += bytesToWrite;
        }

        if (tPtr < tEnd)
        {
            // argh! gotta shrink the target buffer to tPtr bytes :(
            
            byte[] targetOld = target;
            target = new byte[tPtr];
            System.arraycopy(targetOld, 0, target, 0, tPtr);
        }
        return target;
    }

    /**
     * Encodes a String into UTF-8. This implementation computes the
     * needed number of output bytes in advance. It is thus heavier on
     * the CPU, but handles memory allocation more efficiently compared
     * to encode1.
     */
    public static byte[] encode2(String s)
        throws Exception
    {
        final int BYTE_MASK = 0xBF;
        final int BYTE_MARK = 0x80;
        final int flags = STRICT_CONVERSION;

        char[] source = s.toCharArray();
        int sPtr = 0, sEnd = source.length;
        int tPtr = 0;

        while (sPtr < sEnd)
        {
            int bytesToWrite = 0;
            int ch = source[sPtr++];
            
            /* If we have a surrogate pair, convert to UTF32 first. */
            
            if (ch >= UNI_SUR_HIGH_START && ch <= UNI_SUR_HIGH_END && sPtr < sEnd)
            {
                int ch2 = source[sPtr];

                if (ch2 >= UNI_SUR_LOW_START && ch2 <= UNI_SUR_LOW_END)
                {
                    ch = ((ch - UNI_SUR_HIGH_START) << HALF_SHIFT)
                        + (ch2 - UNI_SUR_LOW_START) + HALF_BASE;
                    ++sPtr;
                }
                else if (flags == STRICT_CONVERSION)
                {
                    /* it's an unpaired high surrogate */
                    --sPtr; /* return to the illegal value itself */
                    throw new Exception("expected a surrogate character at position " + sPtr);
                }
            }
            else if ((flags == STRICT_CONVERSION) && (ch >= UNI_SUR_LOW_START && ch <= UNI_SUR_LOW_END))
            {
                --sPtr; /* return to the illegal value itself */
                throw new Exception("expected a surrogate character at position " + sPtr);
            }
            
            /* Figure out how many bytes the result will require */
            
            if (ch < 0x80) bytesToWrite = 1;
            else if (ch < 0x800) bytesToWrite = 2;
            else if (ch < 0x10000) bytesToWrite = 3;
            else if (ch < 0x200000) bytesToWrite = 4;
            else
            {
                bytesToWrite = 2;
                ch = UNI_REPLACEMENT_CHAR;
            }

            tPtr += bytesToWrite;
        }

        byte[] target = new byte[tPtr];
        int tEnd = tPtr;
        tPtr = 0;
        sPtr = 0; sEnd = source.length;

        while (sPtr < sEnd)
        {
            int bytesToWrite = 0;
            int ch = source[sPtr++];
            
            /* If we have a surrogate pair, convert to UTF32 first. */
            
            if (ch >= UNI_SUR_HIGH_START && ch <= UNI_SUR_HIGH_END && sPtr < sEnd)
            {
                int ch2 = source[sPtr];

                if (ch2 >= UNI_SUR_LOW_START && ch2 <= UNI_SUR_LOW_END)
                {
                    ch = ((ch - UNI_SUR_HIGH_START) << HALF_SHIFT)
                        + (ch2 - UNI_SUR_LOW_START) + HALF_BASE;
                    ++sPtr;
                }
                else if (flags == STRICT_CONVERSION)
                {
                    /* it's an unpaired high surrogate */
                    --sPtr; /* return to the illegal value itself */
                    throw new Exception("expected a surrogate character at position " + sPtr);
                }
            }
            else if ((flags == STRICT_CONVERSION) && (ch >= UNI_SUR_LOW_START && ch <= UNI_SUR_LOW_END))
            {
                --sPtr; /* return to the illegal value itself */
                throw new Exception("expected a surrogate character at position " + sPtr);
            }
            
            /* Figure out how many bytes the result will require */
            
            if (ch < 0x80) bytesToWrite = 1;
            else if (ch < 0x800) bytesToWrite = 2;
            else if (ch < 0x10000) bytesToWrite = 3;
            else if (ch < 0x200000) bytesToWrite = 4;
            else
            {
                bytesToWrite = 2;
                ch = UNI_REPLACEMENT_CHAR;
            }

            tPtr += bytesToWrite;
            
            if (tPtr > tEnd)
            {
                byte[] targetOld = target;
                target = new byte[Math.min(tPtr * 2, source.length * 4)];
                System.arraycopy(targetOld, 0, target, 0, tPtr-bytesToWrite);
                tEnd = target.length;
            }
            switch (bytesToWrite)
            {    /* note: everything falls through. */
                case 4:    target[--tPtr] = (byte)((ch | BYTE_MARK) & BYTE_MASK); ch >>= 6;
                case 3:    target[--tPtr] = (byte)((ch | BYTE_MARK) & BYTE_MASK); ch >>= 6;
                case 2:    target[--tPtr] = (byte)((ch | BYTE_MARK) & BYTE_MASK); ch >>= 6;
                case 1:    target[--tPtr] = (byte)(ch | FIRST_BYTE_MARK[bytesToWrite]);
            }
            tPtr += bytesToWrite;
        }

        if (tPtr < tEnd) throw new Exception("can't be");
        return target;
    }
}
